#include "config.h"
#include "common.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <regex.h>
#include <signal.h>
#ifdef HAVE_LIBUTIL_H
#include <libutil.h>
#endif
#if (WIN32|WINNT)
#include <time.h>
#else
#include <sys/time.h>
#endif
#include "log.h"
#include "indexer.h"
#include "robots.h"
#include "dbms.h"
#include "parseurl.h"
#include "charset.h"
#include "parser.h"
#include "spell.h"
#include "md5.h"
#include "proto.h"
#define MD5_DIGEST(data, size, digest) md5_buffer_str(data, size, digest)


/**********************************/


typedef struct filter_struct {
	regex_t filter;
	int     filter_type;
	int	reverse;
	char	*regstr;
} UDM_FILTER;


/* MULTI THREAD variables */
static SERVER		*	Server=NULL;
static UDM_FILTER	*	Filter=NULL;
static ALIAS		*	Alias=NULL;
static int nservers=0,mservers=0;
static int nfilters=0,mfilters=0;
static int naliases=0,maliases=0;
static char ExtraHeaders[STRSIZ]="";
static char UserAgent[STRSIZ]="";
static int MaxDocSize=MAXDOCSIZE;

/***********************************************************/

static int get_gmtoff()
{
#ifdef HAVE_TM_GMTOFF
	struct tm *tim;time_t clock;
	clock=time(0);tim=localtime(&clock);
	return(tim->tm_gmtoff);
#else
	time_t tt = time(NULL);
	struct tm gmt;
	struct tm *t;
	int days, hours, minutes;

	gmt = *gmtime(&tt);
	t = localtime(&tt);
	days = t->tm_yday - gmt.tm_yday;
	hours = ((days < -1 ? 24 : 1 < days ? -24 : days * 24)
		+ t->tm_hour - gmt.tm_hour);
	minutes=hours * 60 + t->tm_min - gmt.tm_min;
	return(minutes*60);
#endif
}


__INDLIB__ int ShowStatistics(INDEXER * Indexer){
int expired_total=0;
int total_total=0;
STAT *stat,*s;


#if (WIN32|WINNT)
#else
	printf("\n          UdmSearch statistics\n\n");
	printf("%10s %10s %10s\n","Status","Expired","Total");
	printf("   -----------------------------\n");
#endif
	stat=GetStatistics(Indexer);
	if(DBErrorCode(Indexer->db)){
		return(IND_ERROR);
	}
	if((s=stat)){
		while(stat->status!=-1){
			SHOWSTAT(Indexer->handle,stat->status,stat->expired,stat->total,
				HTTPErrMsg(stat->status));
			total_total+=stat->total;
			expired_total+=stat->expired;
			stat++;
		}
		free(s);
	}
#if (WIN32|WINNT)
		SHOWSTAT(Indexer->handle,-1,expired_total,total_total,"");
#else
	printf("   -----------------------------\n");
	printf("%10s %10d %10d","Total",expired_total,total_total);
	printf("\n\n");
#endif
	return(IND_OK);
}

static int cmpserver(const void *s1,const void *s2)
{
	return(strlen(((SERVER*)s2)->url)-strlen(((SERVER*)s1)->url));
}

static void FreeFilters(){
int i;
	for(i=0; i < nfilters; i++){
		regfree(&(Filter[i].filter));
		free(Filter[i].regstr);
	}
	free_it(Filter);
	nfilters=0;
	mfilters=0;
}

static int AddFilter(char * filter,int filter_type,int reverse){
#define ERRSTRSIZE 100
char regerrstr[STRSIZ]="";
int err;
	if(nfilters>=mfilters){
		if(mfilters){
			mfilters+=16;
			Filter=(UDM_FILTER *)realloc(Filter,mfilters*sizeof(UDM_FILTER));
		}else{
			mfilters=16;
			Filter=(UDM_FILTER *)malloc(mfilters*sizeof(UDM_FILTER));
		}
	}
	err=regcomp(&(Filter[nfilters].filter),filter,REG_EXTENDED|REG_ICASE);
	if(err){
		regerror(err, &(Filter[nfilters].filter), regerrstr, ERRSTRSIZE);
		udm_log(0,UDM_LOG_ERROR,"Wrong regex in config file: %s: %s", filter,regerrstr);
		regfree(&(Filter[nfilters].filter));
		return(1);
	}
	Filter[nfilters].filter_type=filter_type;
	Filter[nfilters].reverse=reverse;
	Filter[nfilters].regstr=strdup(filter);
	nfilters++;
	return(0);
}

static char FCBuf[64]="";
static char *FilterComStr(UDM_FILTER* f){
	FCBuf[0]=0;
	if(!f)return(FCBuf);
	switch(f->filter_type){
		case F_ALLOW: sprintf(FCBuf,"Allow");break;
		case F_DISALLOW: sprintf(FCBuf,"Disallow");break;
		case F_HEAD: sprintf(FCBuf,"CheckOnly");break;
		default: sprintf(FCBuf,"Unknown");break;
	}
	sprintf(STREND(FCBuf),"%s %s",
		(f->reverse?"NoMatch":""),
		(f->regstr?f->regstr:""));
	return(FCBuf);
}

/* Allow/Disallow/CheckOnly */
static UDM_FILTER * WhatToDo(char *url){
int i;
#define NS 10
regmatch_t subs[NS];
int err;
	for(i=0;i<nfilters;i++){
		err=regexec(&(Filter[i].filter),url,NS,subs,0);
		if((Filter[i].reverse)&&(err))
			return(&Filter[i]);
		if((!Filter[i].reverse)&&(!err))
			return(&Filter[i]);
	}
	return(NULL);
}
static int AddAlias(char * find,char *replace){
	if(naliases>=maliases){
		if(maliases){
			maliases+=16;
			Alias=(ALIAS *)realloc(Alias,maliases*sizeof(ALIAS));
		}else{
			maliases=16;
			Alias=(ALIAS *)malloc(maliases*sizeof(ALIAS));
		}
	}
	Alias[naliases].find=strdup(find);
	Alias[naliases].replace=strdup(replace);
	naliases++;
	return(0);
}
static void FreeAliases(){
int i;
	for(i=0;i<naliases;i++){
		free(Alias[i].find);
		free(Alias[i].replace);
	}
	naliases=0;
	maliases=0;
	free_it(Alias);
}

static int AddServer(SERVER * srv){

	if(nservers>=mservers){
		if(mservers){
			mservers+=16;
			Server=(SERVER *)realloc(Server,mservers*sizeof(SERVER));
		}else{
			mservers=16;
			Server=(SERVER *)malloc(mservers*sizeof(SERVER));
		}
	}
	Server[nservers].url=strdup(srv->url);
	Server[nservers].charset=srv->charset?strdup(srv->charset):NULL;
	Server[nservers].basic_auth=srv->basic_auth?strdup(srv->basic_auth):NULL;
	Server[nservers].htdb_list=srv->htdb_list?strdup(srv->htdb_list):NULL;
	Server[nservers].htdb_doc=srv->htdb_doc?strdup(srv->htdb_doc):NULL;
	if(srv->proxy){
		char *s;
		Server[nservers].proxy=strdup(srv->proxy);
		if((s=strchr(Server[nservers].proxy,':'))){
			*s=0;
			Server[nservers].proxy_port=atoi(s+1)?atoi(s+1):DEFAULT_PROXY_PORT;
		}else{
			Server[nservers].proxy_port=DEFAULT_PROXY_PORT;
		}
	}else{
		Server[nservers].proxy=NULL;
		Server[nservers].proxy_port=0;
	}
	Server[nservers].period=srv->period;
	Server[nservers].hint=srv->hint;
	Server[nservers].net_errors=0;
	Server[nservers].outside=srv->outside;
	Server[nservers].maxhops=srv->maxhops;
	Server[nservers].gindex=srv->gindex;
	Server[nservers].gfollow=srv->gfollow;
	Server[nservers].deletebad=srv->deletebad;
	Server[nservers].userobots=srv->userobots;
	Server[nservers].bodyweight=srv->bodyweight;
	Server[nservers].titleweight=srv->titleweight;
	Server[nservers].urlweight=srv->urlweight;
	Server[nservers].descweight=srv->descweight;
	Server[nservers].keywordweight=srv->keywordweight;
	Server[nservers].max_net_errors=srv->max_net_errors;
	Server[nservers].read_timeout=srv->read_timeout;
	nservers++;
	return(0);
}

static void FreeServers(){
int i;
	for(i=0;i<nservers;i++){
		free_it(Server[i].url);
		free_it(Server[i].charset);
		free_it(Server[i].basic_auth);
		free_it(Server[i].htdb_list);
		free_it(Server[i].htdb_doc);
		free_it(Server[i].proxy);
	}
	nservers=mservers=0;
	free_it(Server);
}

static SERVER * FindServer(char *url){
int i;
char *r;
	if(!Server)return(NULL);
	url=remove2dot(url);
	r=strstr(url,"/robots.txt");
	/* If it's a robot.txt, cut to hostinfo and find result */
	if((r)&&(!strcmp(r,"/robots.txt"))){
		r++;*r=0;
		for(i=0;i<nservers;i++){
			if(!STRNCMP(url,Server[i].url)){
				*r='r';
				return(&Server[i]);
			}
		}
		*r='r';
	}else{
		for(i=0;i<nservers;i++)
			if(!STRNCMP(url,Server[i].url))
				return(&Server[i]);
	}
	return(0);
}

#define RESORT_WORDS 256
static int cmpwords(const void * v1, const void * v2){
	return(strcmp(((UDM_WORD*)v1)->word,((UDM_WORD*)v2)->word));
}
static int AddWord1(INDEXER *Indexer,char *word,int where,int checkstop){
int i;
char *s;
STOP_WORD * stop;

int l,r,c,res;

	s=strdup(word);
	if(strlen(s)>MAXWORDSIZE)s[MAXWORDSIZE]=0;
	tolower_string(s,LCindex);
	if(checkstop){
		/* Stopwords checking and language guesser */
		if((stop=IsStopWord(s))){
			int i;

			udm_log(Indexer->handle,UDM_LOG_DEBUG,"stop: '%s'-'%s'", stop->word, stop->lang);
			/* Add language */
			if(!strcmp(stop->lang,""))
				Indexer->lang[0].count++;
			else
			for(i=1;i<MAXLANGPERDOC;i++){
				if(!strcmp(Indexer->lang[i].lang,stop->lang)){
					Indexer->lang[i].count++;
					break;
				}
				/* if current lang is free */
				if(!strcmp(Indexer->lang[i].lang,"")){
					strcpy(Indexer->lang[i].lang,stop->lang);
					Indexer->lang[i].count++;
					break;
				}
			}
			free(s);
			return(0);
		}
	}
	
	/* Find current word in sorted part of word list */
	l=0;r=Indexer->swords-1;
	while(l<=r){
		c=(l+r)/2;
		res=strcmp(Indexer->Word[c].word,s);
		if(res==0){
			Indexer->Word[c].count|=where;
			free(s);
			return(0);
		}
		if(res<0)
			l=c+1;
		else
			r=c-1;
	}
	/* Find in unsorted part */
	for(i=Indexer->swords;i<Indexer->nwords;i++){
		if(!strcmp(Indexer->Word[i].word,s)){
			Indexer->Word[i].count|=where;
			free(s);return(0);
		}
	}
	/* Add new word */
	if(Indexer->nwords>=Indexer->mwords){
		if(Indexer->mwords){
			Indexer->mwords+=WSIZE;
			Indexer->Word=(UDM_WORD *)realloc(Indexer->Word,Indexer->mwords*sizeof(UDM_WORD));
		}else{
			Indexer->mwords=WSIZE;
			Indexer->Word=(UDM_WORD *)malloc(Indexer->mwords*sizeof(UDM_WORD));
		}
	}
	Indexer->Word[Indexer->nwords].word=s;
	Indexer->Word[Indexer->nwords].count=where;
	Indexer->Word[Indexer->nwords].code=0;
	Indexer->nwords++;
	
	/* Sort unsorted part */
	if((Indexer->nwords-Indexer->swords)>RESORT_WORDS){
		qsort(Indexer->Word,Indexer->nwords,sizeof(UDM_WORD),cmpwords);
		Indexer->swords=Indexer->nwords;
	}
	return(0);
}

static int AddCode(INDEXER *Indexer,int code,int where){
int i;

	for(i=0;i<Indexer->nwords;i++){
		if(Indexer->Word[i].code==code){
			Indexer->Word[i].count|=where;
			return(0);
		}
	}
	if(Indexer->nwords>=Indexer->mwords){
		if(Indexer->mwords){
			Indexer->mwords+=WSIZE;
			Indexer->Word=(UDM_WORD *)realloc(Indexer->Word,Indexer->mwords*sizeof(UDM_WORD));
		}else{
			Indexer->mwords=WSIZE;
			Indexer->Word=(UDM_WORD *)malloc(Indexer->mwords*sizeof(UDM_WORD));
		}
	}
	Indexer->Word[Indexer->nwords].word=strdup("");
	Indexer->Word[Indexer->nwords].count=where;
	Indexer->Word[Indexer->nwords].code=code;
	Indexer->nwords++;return(0);
}



static int AddWord(INDEXER *Indexer,char *word,int where,int checkstop){
char 	** forms, ** saveforms;

	tolower_string(word,LCindex);
	if((saveforms=forms=NormalizeWord(word))){
		/* Add all NORMAL forms of the word */
		while(*forms){
			AddWord1(Indexer,*forms,where,checkstop);
			free(*forms);
			forms++;
		}
		free(saveforms);
	}else{
		/* If NORMAL forms has not been found */
		/*   then we will add the word itself */
		AddWord1(Indexer,word,where,checkstop);
	}
	return(0);
}


static int AddWordWithCode(INDEXER *Indexer,char *word,int where,int checkstop){
int * forms, * saveforms;

	tolower_string(word,LCindex);
	if((saveforms=forms=IntNormalizeWord(word))){
		/* Add all NORMALized forms of the word */
		while(*forms){
			AddCode(Indexer,*forms,where);
			forms++;
		}
		free(saveforms);
	}else{
		/* If NORMALized forms have not been found */
		/*   then we will add the word itself */
		AddWord1(Indexer,word,where,checkstop);
	}
	return(0);
}


static int FreeWords(INDEXER* Indexer) {
	int i;
	for(i=0;i<Indexer->nwords;i++) free(Indexer->Word[i].word);
	Indexer->nwords=0;
	return(0);
}

int InitServer(SERVER * srv){
	srv->url=NULL;
	srv->proxy=NULL;
	srv->basic_auth=NULL;
	srv->charset=NULL;
	srv->htdb_list=NULL;
	srv->htdb_doc=NULL;
	srv->proxy_port=0;
	srv->period=DEFAULT_REINDEX_TIME;
	srv->hint=0;
	srv->net_errors=0;
	srv->max_net_errors=MAXNETERRORS;
	srv->read_timeout=READ_TIMEOUT;
	srv->outside=0;
	srv->maxhops=DEFAULT_MAX_HOPS;
	srv->gindex=1;
	srv->gfollow=1;
	srv->deletebad=0;
	srv->userobots=1;
	srv->bodyweight=1;
	srv->titleweight=2;
	srv->keywordweight=2;
	srv->descweight=2;
	srv->urlweight=0;	
	return(0);
}

/* One server structure to be available on all config levels*/
static SERVER * csrv=NULL;

__INDLIB__ int LoadConfig(char *conf_name,int config_level,int load_flags){
int i=0; /* For line number */
FILE *config;
char str[STRSIZ]="";
char str1[STRSIZ]="";
char str2[STRSIZ]="";
int has_dict=0;

	
	if(config_level==0){ /* Do some initialization */
		csrv=(SERVER*)malloc(sizeof(SERVER));
		InitServer(csrv);
		SetMaxURLNumber(-1);
		CurURLNumber=0;
		sprintf(UserAgent,"%s/%s",INDEXER_USER_AGENT,VERSION);
		ExtraHeaders[0]=0;
		MaxDocSize=MAXDOCSIZE;
		GMToffs=get_gmtoff();
#ifdef HAVE_SYSLOG_H
		LogFacility=LOG_FACILITY;
#endif
		Force1251=0;
		DBPort=0;
		free_it(DBHost);
		free_it(DBName);
		free_it(DBUser);
		free_it(DBPass);
		free_it(LocalCharset);
		LCindex=0;
		makeseparator(SEPARATOR,LCindex);
		FreeHrefs();
		FreeFilters();
		FreeAliases();
		free_it(MirrorRoot);
		free_it(MirrorHeadersRoot);
		use_mirror = UDM_MIRROR_NO;
#ifdef USE_FILE
		FreeTypes();
#endif
		FreeServers();
		have_targets=1;
		index_started=0;
		mark_done=!(load_flags&FLAG_MARK);
	}

	/* Open config file */
	if(!(config=fopen(conf_name,"r"))){
		udm_log(0,UDM_LOG_ERROR,"Error: can't open config file '%s': %s",conf_name, strerror(errno));
		return(1);
	}

	/*  Read lines and parse */
	while(fgets(str1,sizeof(str1),config)){
		char *end;
		i++;
		if(!str1[0])continue;
		end=str1+strlen(str1)-1;
		while((end>=str1)&&(*end=='\r'||*end=='\n')){
			*end=0;if(end>str1)end--;
		}
		if(!str1[0])continue;
		if(str1[0]=='#')continue;

		if(*end=='\\'){
			*end=0;strcat(str,str1);
			continue;
		}
		strcat(str,str1);
		strcpy(str1,"");

		if(!STRNCASECMP(str,"DBHost")){
			if(sscanf(str+6,"%s",str1))
				DBHost=strdup(str1);
		}else
		if(!STRNCASECMP(str,"DBName")){
			DBName=strdup(trim_chars(str+6," \r\n\t"));
		}else
		if(!STRNCASECMP(str,"DBUser")){
			if(sscanf(str+6,"%s",str1)&&strcmp(str1,""))
				DBUser=strdup(str1);
		}else
		if(!STRNCASECMP(str,"DBPass")){
			if(sscanf(str+6,"%s",str1)&&strcmp(str1,""))
				DBPass=strdup(str1);
			else	DBPass=strdup("");
		}else
		if(!STRNCASECMP(str,"DBPort")){
			if(sscanf(str+6,"%d",&DBPort)!=1)DBPort=0;
		}else
		if(!STRNCASECMP(str,"LocalCharset")){
			if(sscanf(str+13,"%s",str1)&&strcmp(str1,"")){
				free_it(LocalCharset);
				LocalCharset=strdup(str1);
				LCindex=getcharset(LocalCharset);
				makeseparator(SEPARATOR,LCindex);
			}
		}else
		if(!STRNCASECMP(str,"DBType")){
			if(sscanf(str+6,"%s",str1)){
				if(!STRNCASECMP(str1,"msql"))DBType=DB_MSQL;
				else
				if(!STRNCASECMP(str1,"solid"))DBType=DB_SOLID;
				else
				if(!STRNCASECMP(str1,"oracle"))DBType=DB_ORACLE;
				else
					DBType = DB_UNK;
			}else{
				DBType = DB_UNK;
			}
		}else
		if(!STRNCASECMP(str,"DisallowNoMatch")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,F_DISALLOW,1))
					return(1);
			}
		}else
		if(!STRNCASECMP(str,"Disallow")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,F_DISALLOW,0))
					return(1);
			}
		}else
		if(!STRNCASECMP(str,"AllowNoMatch")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,F_ALLOW,1))
					return(1);
			}
		}else
		if(!STRNCASECMP(str,"Allow")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,F_ALLOW,0))
					return(1);
			}
		}else
		if(!STRNCASECMP(str,"CheckOnlyNoMatch")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,F_HEAD,1))
					return(1);
			}
		}else
		if(!STRNCASECMP(str,"CheckOnly")){
			char *s,*lt;
			s=gettoken(str," \t\r\n",&lt);
			while((s=gettoken(NULL," \t\r\n",&lt))){
				if(AddFilter(s,F_HEAD,0))
					return(1);
			}
		}else
#ifdef USE_FILE
		if(!STRNCASECMP(str,"AddType")){
			char *s,*s1,*lt;
			s=gettoken(str," \t\r\n",&lt);
			if((s1=gettoken(NULL," \t\r\n",&lt)))
				while((s=gettoken(NULL," \t\r\n",&lt)))
					AddType(s1,s);
		}else
#endif
		if(!STRNCASECMP(str,"Tag")){
			if(sscanf(str+4,"%d",&csrv->hint)!=1)
				csrv->hint=0;
		}else
		if(!STRNCASECMP(str,"MaxNetErrors")){
			if(sscanf(str+12,"%d",&csrv->max_net_errors)!=1)
				csrv->max_net_errors=MAXNETERRORS;
		}else
		if(!STRNCASECMP(str,"ReadTimeOut")){
			if(sscanf(str+11,"%d",&csrv->read_timeout)!=1)
				csrv->read_timeout=READ_TIMEOUT;
		}else
		if(!STRNCASECMP(str,"Period")){
			if(sscanf(str+6,"%d",&csrv->period)!=1)
				csrv->period=DEFAULT_REINDEX_TIME;
		}else
		if(!STRNCASECMP(str,"CharSet")){
			free_it(csrv->charset);
			if(sscanf(str+7,"%s",str1))
				csrv->charset=strdup(str1);
		}else
		if(!STRNCASECMP(str,"Proxy")){
			free_it(csrv->proxy);
			if(sscanf(str+5,"%s",str1)&&strcmp(str1,""))
				csrv->proxy=strdup(str1);
		}else
		if(!STRNCASECMP(str,"HTTPHeader")){
			if(sscanf(str+11,"%[^\n\r]s",str1)&&strcmp(str1,"")){
				if(!STRNCMP(str1,"User-Agent: ")){
					strcpy(UserAgent,str1+12);
				}else{
					strcat(ExtraHeaders,str1);
					strcat(ExtraHeaders,"\r\n");
				}
			}
		}else
#ifdef HAVE_SYSLOG_H
		if(!STRNCASECMP(str,"SyslogFacility")){
			if(sscanf(str+14,"%s",str1)&&strcmp(str1,""))
				LogFacility=syslog_facility(str1);
		}else
#endif
		if(!STRNCASECMP(str,"AuthBasic")){
			free_it(csrv->basic_auth);
			if(sscanf(str+9,"%s",str1)){
				csrv->basic_auth=(char*)malloc(BASE64_LEN(strlen(str1)));
				base64_encode(str1,csrv->basic_auth,strlen(str1));
			}
		}else
		if(!STRNCASECMP(str,"HTDBList")){
			free_it(csrv->htdb_list);
			csrv->htdb_list=strdup(trim_chars(str+8," \t\r\n"));
		}else
		if(!STRNCASECMP(str,"HTDBDoc")){
			free_it(csrv->htdb_doc);
			csrv->htdb_doc=strdup(trim_chars(str+8," \t\r\n"));
		}else
		if(!STRNCASECMP(str,"Mime")){
			AddMime(str+4);
		}else
		if(!STRNCASECMP(str,"Alias")){
			if(sscanf(str+5,"%s%s",str1,str2)==2)
				AddAlias(str1,str2);
			else{
				udm_log(0,UDM_LOG_ERROR,
					"Error in config file '%s' line %d:%s",
					conf_name,i,str);
				return(1);
			}
		}else
		if(!STRNCASECMP(str,"Server")){
			URL from;
			if(sscanf(str+6,"%s",str1)!=1)continue;

			if(parseurl(&from,str1)){
				udm_log(0,UDM_LOG_ERROR,
					"File '%s' line %d: Invalid URL: %s",
					conf_name,i,str1);
				return(1);
			}
			csrv->url=strdup(str1);
			AddServer(csrv);
			if((!strcmp(from.schema,"http"))&&(load_flags&FLAG_ADD_SERV)&&(csrv->userobots)){
				sprintf(str,"%s://%s/%s",from.schema,from.hostinfo,"robots.txt");
				AddHref(NULL,str,0,0,0);
			}
			if(load_flags&FLAG_ADD_SERV)
				AddHref(NULL,str1,0,0,0);
		}else
		if(!STRNCASECMP(str,"FollowOutside")){
			if(sscanf(str+13,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					csrv->outside=1;
				else
					csrv->outside=0;
			}else
				csrv->outside=0;
		}else
		if(!STRNCASECMP(str,"Index")){
			if(sscanf(str+5,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					csrv->gindex=1;
				else
					csrv->gindex=0;
			}else
				csrv->gindex=1;
		}else
		if(!STRNCASECMP(str,"Follow ")){
			if(sscanf(str+7,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					csrv->gfollow=1;
				else
					csrv->gfollow=0;
			}else
				csrv->gfollow=1;
		}else
		if(!STRNCASECMP(str,"Robots")){
			if(sscanf(str+6,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					csrv->userobots=1;
				else
					csrv->userobots=0;
			}else
				csrv->userobots=1;
		}else
		if(!STRNCASECMP(str,"DeleteBad")){
			if(sscanf(str+9,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))
					csrv->deletebad=1;
				else
					csrv->deletebad=0;
			}else
				csrv->deletebad=0;
		}else
		if(!STRNCASECMP(str,"MaxHops")){
			if(sscanf(str+7,"%d",&csrv->maxhops)!=1)
				csrv->maxhops=0;
		}else
		if(!STRNCASECMP(str,"BodyWeight")){
			if(sscanf(str+10,"%d",&csrv->bodyweight)!=1)
				csrv->bodyweight=1;
		}else
		if(!STRNCASECMP(str,"TitleWeight")){
			if(sscanf(str+11,"%d",&csrv->titleweight)!=1)
				csrv->titleweight=2;
		}else
		if(!STRNCASECMP(str,"DescWeight")){
			if(sscanf(str+10,"%d",&csrv->descweight)!=1)
				csrv->descweight=2;
		}else
		if(!STRNCASECMP(str,"UrlWeight")){
			if(sscanf(str+9,"%d",&csrv->urlweight)!=1)
				csrv->urlweight=0;
		}else
		if(!STRNCASECMP(str,"KeywordWeight")){
			if(sscanf(str+13,"%d",&csrv->keywordweight)!=1)
				csrv->keywordweight=2;
		}else
		if(!STRNCASECMP(str,"MaxDocSize")){
			if(sscanf(str+11,"%d",&MaxDocSize)!=1)
				MaxDocSize=MAXDOCSIZE;
		}else
		if(!STRNCASECMP(str,"NoIndex")){
			csrv->gindex=0;
		}else
		if(!STRNCASECMP(str,"NoFollow")){
			csrv->gfollow=0;
		}else
		if(!STRNCASECMP(str,"ForceIISCharset1251")){
			if(sscanf(str+20,"%s",str1)==1){
				if(!STRNCASECMP(str1,"yes"))Force1251=1;
				else	Force1251=0;
			}else	Force1251=0;
		}else
		if(!STRNCASECMP(str,"Affix")){
			char lang[20];
			if(load_flags&&FLAG_SPELL)
			if(sscanf(str+5,"%s%s",lang,str1)){
				if(*str1=='/')strcpy(str,str1);
				else	sprintf(str,"%s/%s",CONFIGDIR,str1);
				if(ImportAffixes(lang,str)){
					udm_log(0,UDM_LOG_ERROR,
					"Can't load affix :%s",str);
					return(1);
				}
			}
		}else
		if(!STRNCASECMP(str,"Spell")){
			char lang[20];
			if(load_flags&&FLAG_SPELL)
			if(sscanf(str+5,"%s%s",lang,str1)){
				if(*str1=='/')strcpy(str,str1);
				else	sprintf(str,"%s/%s",CONFIGDIR,str1);
				/* -----  0 or 1 ?  */
				if(ImportDictionary(lang,str,0,"")){
					udm_log(0,UDM_LOG_ERROR,
					"Can't load dictionary :%s",str);
					return(1);
				}
				has_dict=1;
			}
		}else
		if(!STRNCASECMP(str,"Include")){
			if((sscanf(str+7,"%s",str1))&&(config_level<5)){
				if(*str1=='/')strcpy(str,str1);
				else	sprintf(str,"%s/%s",CONFIGDIR,str1);
				LoadConfig(str,config_level+1,load_flags);
			}
		}else
		if(!STRNCASECMP(str,"MirrorRoot")){
			if((sscanf(str+10,"%s",str1))){
				if(*str1=='/')
					strcpy(str,str1);
				else
					sprintf(str,"%s/%s",CONFIGDIR,str1);
			} else {
				sprintf(str,"%s/mirrors",CONFIGDIR);
			}
			MirrorRoot = strdup(str);
			use_mirror=(use_mirror>0)?use_mirror:0;
		}else
		if(!STRNCASECMP(str,"MirrorHeadersRoot")){
			if((sscanf(str+17,"%s",str1))){
				if(*str1=='/')
					strcpy(str,str1);
				else
					sprintf(str,"%s/%s",CONFIGDIR,str1);
			} else {
				sprintf(str,"%s/headers",CONFIGDIR);
			}
			MirrorHeadersRoot = strdup(str);
			use_mirror=(use_mirror>0)?use_mirror:0;
		}else
		if(!STRNCASECMP(str, "MirrorPeriod")){
			/* May be better make it useful for every Server? */
			if(sscanf(str + 12, "%d", &use_mirror) != 1)
				use_mirror = UDM_MIRROR_NO;
		}else{
			udm_log(0,UDM_LOG_ERROR,"Error in config file line %d: %s",i,str);
			return(1);
		}
		str[0]=0;
	}
	fclose(config);


	/* On level0 : Free some variables, prepare others, etc */
	if(config_level==0){
		/* Add one virtual server if we want to follow outside */
		if(csrv->outside){
			free_it(csrv->url);
			csrv->url="";
			AddServer(csrv);
		}else{
			free_it(csrv->url);
		}
		free_it(csrv->charset);
		free_it(csrv->proxy);
		free_it(csrv->basic_auth);
		free_it(csrv->htdb_list);
		free_it(csrv->htdb_doc);
		/*  Long name should be found first    */
		/*  to allow different options         */
		/*  for server and it's subdirectories */
		qsort((void*)Server,nservers,sizeof(SERVER),cmpserver);
		/* Sort ispell dictionay if it has been loaded */
		if(has_dict)SortDictionary();
	}
	return(0);
}
static int ParseText(INDEXER * Indexer,char *content,char *separator,int weight,int check){
char *s,*lt;
	if(weight&&content&&separator){
		s=gettoken(content,separator,&lt);
		while(s){
			AddWord(Indexer,s,weight,check);
			s=gettoken(NULL,separator,&lt);
		}
	}return(0);
}
static int ParseHtml(INDEXER * Indexer,char *content,char *separator,URL * curURL,URL * realURL,
DOCUMENT * doc,SERVER * CurSrv,int index,int follow,
char *text,char *keywords, char *descript,char *title){

int len,inbody,inscript,intitle,i,j,state,Method;
char sch=0,*s=0,*e=0,*tmp=0,*href=0,*newschema=0,*lt;
URL newURL;TAG tag;
UDM_FILTER *MatchedFilter=NULL;
char str[STRSIZ]="";
char str1[STRSIZ]="";

	inbody=inscript=intitle=0;state=HTML_UNK;s=content;
	while(*s){
		switch(state){
		case HTML_UNK: /* yet unknown */
			if(!STRNCMP(s,"<!--"))
				state=HTML_COM;
			else
			if(*s=='<')
				state=HTML_TAG;
			else
				state=HTML_TXT;
			break;

		case HTML_COM: /* comment */
			e=strstr(s,"-->");
			if(e)len=e-s+3;
			else len=strlen(s);
			s=s+len; state=HTML_UNK;
			break;

		case HTML_TAG: /* tag */
			e=strchr(s,'>');
			if(e)len=e-s+1;
			else len=strlen(s);
			tmp=(char*)malloc(len+1);
			strncpy(tmp,s,len);tmp[len]=0;
			parsetag(&tag,tmp);

			if(!strcasecmp(tag.name,"body"))inbody=1;
			else
			if(!strcasecmp(tag.name,"/body"))inbody=0;
			else
			if(!strcasecmp(tag.name,"script"))inscript=1;
			else
			if(!strcasecmp(tag.name,"/script"))inscript=0;
			else
			if(!strcasecmp(tag.name,"title"))intitle=1;
			else
			if(!strcasecmp(tag.name,"/title"))intitle=0;
			else
			if((!strcasecmp(tag.name,"a"))&&(tag.href))
				href=strdup(tag.href);
			else
			if((!strcasecmp(tag.name,"area"))&&(tag.href))
				href=strdup(tag.href);
			else
			if((!strcasecmp(tag.name,"link"))&&(tag.href))
				href=strdup(tag.href);
			else
			if((!strcasecmp(tag.name,"frame"))&&(tag.src))
				href=strdup(tag.src);
			else
			if((!strcasecmp(tag.name,"img"))&&(tag.src))
				href=strdup(tag.src);
			else
			if((!strcasecmp(tag.name,"meta"))&&(tag.equiv)&&(tag.content)){
				if((!strcasecmp(tag.equiv,"Content-Type"))&&(!Indexer->charset)){
					if((Indexer->charset=strstr(tag.content,"charset="))){
						Indexer->charset+=8;
						recode_string(content,getcharset(Indexer->charset),LCindex);
					}
				}else
				if(!strcasecmp(tag.equiv,"refresh")){
					char *u;
					if((u=strstr(tag.content,"URL=")))
						href=strdup(u+4);
				}else
				if(!strcasecmp(tag.equiv,"keywords")){
					strncpy(keywords,tag.content,MAXKEYWORDSIZE-1);
					keywords[MAXKEYWORDSIZE-1]=0;
					if(index&&CurSrv->keywordweight){
						strcpy(str,keywords);
						ParseText(Indexer,str,separator,
							CurSrv->keywordweight,1);
					}
				}else
				if(!strcasecmp(tag.equiv,"description")){
					strncpy(descript,tag.content,MAXDESCSIZE-1);
					descript[MAXDESCSIZE-1]=0;
					if(index&&CurSrv->descweight){
						strcpy(str,descript);
						ParseText(Indexer,str,separator,
							CurSrv->descweight,1);
					}
				}else
				if(!strcasecmp(tag.equiv,"robots")){
					if(CurSrv->userobots){
					e=gettoken(tag.content,separator,&lt);
					while(e){
						if(!strcasecmp(e,"ALL")){
							follow=1;index=1;
						}else
						if(!strcasecmp(e,"NONE")){
							follow=0;index=0;
						}else
						if(!strcasecmp(e,"NOINDEX"))
							index=0;
						else
						if(!strcasecmp(e,"NOFOLLOW"))
							follow=0;
						else
						if(!strcasecmp(e,"INDEX"))
							index=1;
						else
						if(!strcasecmp(e,"FOLLOW"))
							follow=1;
						e=gettoken(NULL,separator,&lt);
					}}
				}
			}
			if((href)&&(follow)&&(CurSrv->gfollow)){
				if(doc->hops>=CurSrv->maxhops){
					udm_log(Indexer->handle,UDM_LOG_DEBUG,"Skip \"%s\" : too many hops: %d",href,doc->hops);
				}else
				if(!parseurl(&newURL,href)){
					if(newURL.schema[0])newschema=newURL.schema;
					else	newschema=curURL->schema;
					if(!strcmp(newschema,"file")||!strcmp(newschema,"htdb")){
						sprintf(str,"%s:%s%s",newschema,newURL.path[0]?newURL.path:curURL->path,newURL.filename);
					}else{
						sprintf(str,"%s://%s%s%s",
						newURL.schema[0]?newURL.schema:curURL->schema,
						newURL.hostinfo[0]?newURL.hostinfo:curURL->hostinfo,
						newURL.path[0]?newURL.path:curURL->path,newURL.filename);
					}
					remove2dot(str);
					if(!STRNCMP(str,"ftp://")&&(strstr(str,";type=")))
						*(strstr(str,";type"))=0;

					if((MatchedFilter=WhatToDo(str)))
						Method=MatchedFilter->filter_type;
					else
						Method=F_ALLOW;

					if((Method==F_DISALLOW)){
						udm_log(Indexer->handle,UDM_LOG_DEBUG,"Skip \"%s\" : '%s'",href,FilterComStr(MatchedFilter));
						udm_log(Indexer->handle,UDM_LOG_DEBUG,"Full URL \"%s\"",str);
					}else
					if((FindRobots(Indexer,str)>=0)&&(CurSrv->userobots)){
						udm_log(Indexer->handle,UDM_LOG_DEBUG,"Skip \"%s\" : robots",href);
						udm_log(Indexer->handle,UDM_LOG_DEBUG,"Full URL \"%s\"",str);
					}else
					if(!FindServer(str)){
						udm_log(Indexer->handle,UDM_LOG_DEBUG,"Skip \"%s\" : no Server",href);
					}else{
						if((!strcmp(newURL.schema,"http"))&&(CurSrv->outside)&&(CurSrv->userobots)){
							sprintf(str1,"%s://%s/%s",newURL.schema,newURL.hostinfo,"robots.txt");
							AddHref(NULL,str1,0,0,0);
						}
						AddHref(NULL,str,doc->url_id,doc->hops+1,0);
					}
					freeurl(&newURL);
				}
				free_it(href);
			}
			freetag(&tag);free(tmp);
			s=s+len; state=HTML_UNK;
			break;

		case HTML_TXT: /* text */
			e=strchr(s,'<');
			if(e) len=e-s;
			else  len=strlen(s);
			tmp=(char*)malloc(2*len+6);
			e=tmp;j=0;*e=0;
			for(i=0;i<len;i++){
				switch(s[i]){
				case ' ' : case '\t':
				case '\n': case '\r':
					j++; break;

				case '&': /* parse specials */
					if(j){*e=' ';e++;j=0;}
					i++;
					if((sch=SgmlToChar(s+i,LCindex))){
						if(sch==' ')j++;
						*e=sch;e++;*e=0;
						while((i<len)&&(isalnum(*(s+i+1))))
							i++;
						if(s[i+1]==';')i++;
					}else{
						*e='?';e++;*e=0;
					}
					break;
				default:
					if(j){*e=' ';e++;j=0;}
					*e=(s[i]=='\'')?'`':s[i];
					e++; *e=0;
					break;
				}
			}
			if((inbody&&!inscript)&&(strlen(text)<MAXTEXTSIZE)){
				if(*text)strcat(text," ");
				strncat(text,tmp,MAXTEXTSIZE);
			}
			if((intitle)&&(strlen(title)<MAXTITLESIZE)){
				if(*title)strcat(title," ");
				strncat(title,tmp,MAXTITLESIZE-1);
				title[MAXDESCSIZE-1]=0;
			}
			if(index&&!inscript){
				ParseText(Indexer,tmp,separator,
					CurSrv->bodyweight*inbody+
					CurSrv->titleweight*intitle,1);
			}
			free(tmp);
			s=s+len; state=HTML_UNK;
			break;

		}
	}
	return(0);
}

static int cmplang(const void *s1,const void *s2)
{
	return(((UDM_LANG*)s2)->count-((UDM_LANG*)s1)->count);
}

__INDLIB__ int IndexNextURL(INDEXER *Indexer,int index_flags){
char str[STRSIZ]="",str1[STRSIZ]="";
char text[2*MAXTEXTSIZE]="";      char text_escaped[2*MAXTEXTSIZE+1]="";
char keywords[MAXKEYWORDSIZE]=""; char keywords_escaped[2*MAXKEYWORDSIZE+1]="";
char descript[MAXDESCSIZE]="";    char descript_escaped[2*MAXDESCSIZE+1]="";
char last_modified[32]="";        char last_modified_escaped[32*2+1]="";
char title[MAXTITLESIZE]="";      char title_escaped[MAXTITLESIZE*2+1];
char subj[MAXTITLESIZE]="";
char from[MAXKEYWORDSIZE]="";	char digest[33];
char *s,*content,*header,*content_type,*location,*href,*statusline,*lt;
int j,size,status,realsize,Method=0;
int origin,clone,follow,index,changed;
int reindex;
int result=IND_UNKNOWN;
int found_in_mirror = 0;
char *hcopy=NULL;

#ifdef USE_PARSER
char *mime;
int mimeno;
#endif

URL curURL,newURL,realURL;
DOCUMENT * Doc;
UDM_FILTER *MatchedFilter=NULL;
SERVER *CurSrv=0;


	for(j=0;j<MAXLANGPERDOC;j++){
		Indexer->lang[j].lang[0]=0;
		Indexer->lang[j].count=0;
	}

	/* Get Next URL to be indexed from the database */
	THREADINFO(Indexer->handle,"Selecting","");
	if(!(Doc=GetDocInfo(Indexer,index_flags))){
		if(DBErrorCode(Indexer->db)){
			return(IND_ERROR);
		}else{
			return(IND_NO_TARGET);
		}
	}
	if(Indexer->action==IND_TERMINATED)return(IND_TERMINATED);
	IncURLNumber();

	reindex=index_flags&FLAG_REINDEX;

	if(!Indexer->buf){
		/* Alloc buffer for document */
		Indexer->buf=(char*)malloc(MaxDocSize);
	}

	udm_log(Indexer->handle,INFO,"%s",Doc->url);
	THREADINFO(Indexer->handle,"Found",Doc->url);

	/* Add it in cache */
	AddHref(Indexer,Doc->url,0,Doc->hops,1);

#ifdef HAVE_SETPROCTITLE
	/* setproctitle("%s",Doc->url); */
#endif
	if(parseurl(&curURL,Doc->url)){
		result=DeleteUrl(Indexer,Doc->url_id);
		udm_log(Indexer->handle,WARNING,"Invalid URL: %s ... deleted",Doc->url);
		FreeDoc(Doc);
		return(result);
	}

	if(!(CurSrv=FindServer(Doc->url))){
		if(!strcmp(curURL.filename,"robots.txt")){
			if(IND_OK==(result=DeleteRobotsFromHost(Indexer,curURL.hostinfo)))
				result=LoadRobots(Indexer);
		}else{
			udm_log(Indexer->handle,WARNING,"No 'Server' command for url... deleted.");
			result=IND_OK;
		}
		if(result==IND_OK)result=DeleteUrl(Indexer,Doc->url_id);
		freeurl(&curURL);
		FreeDoc(Doc);
		return(result);
	}

	if(Doc->hops>CurSrv->maxhops){
		result=DeleteUrl(Indexer,Doc->url_id);
		udm_log(Indexer->handle,WARNING,"Too many hops (%d)... deleted.",Doc->hops);
		freeurl(&curURL);
		FreeDoc(Doc);
		return(result);
	}
	if(!(MatchedFilter=WhatToDo(Doc->url)))
		Method=F_ALLOW;
	else
		Method=MatchedFilter->filter_type;

	if(Method==F_DISALLOW){
		udm_log(Indexer->handle,WARNING,"Deleting URL : '%s'",FilterComStr(MatchedFilter));
		result=DeleteUrl(Indexer,Doc->url_id);
		if((result==IND_OK)&&(!strcmp(curURL.filename,"robots.txt"))){
			if(IND_OK==(result=DeleteRobotsFromHost(Indexer,curURL.hostinfo)))
				result=LoadRobots(Indexer);
		}
		freeurl(&curURL);
		FreeDoc(Doc);
		return(result);
	}

	if((CurSrv->userobots)&&(FindRobots(Indexer,Doc->url)>=0)){
		udm_log(Indexer->handle,WARNING,"Deleting URL: robots.txt");
		result=DeleteUrl(Indexer,Doc->url_id);
	}else
	if((!strcmp(curURL.schema,"ftp"))&&(!CurSrv->proxy)){
		udm_log(Indexer->handle,WARNING,"FTP is available via proxy only, skipped");
		result=UpdateUrl(Indexer,Doc->url_id,504,CurSrv->period);
	}else
	if(CurSrv->net_errors>CurSrv->max_net_errors){
		udm_log(Indexer->handle,WARNING,"Too many network errors for this server, skipped");
		result=UpdateUrl(Indexer,Doc->url_id,504,CurSrv->period);
	}
	if(result){
		freeurl(&curURL);
		FreeDoc(Doc);
		return(result);
	}


	/* Find alias */
	parseurl(&realURL,Doc->url);
	for(j=0;j<naliases;j++){
		if(!STRNCMP(Doc->url,Alias[j].find)){
			sprintf(str1,"%s%s",Alias[j].replace,Doc->url+strlen(Alias[j].find));
			freeurl(&realURL);
			if(parseurl(&realURL,str1)){
				udm_log(Indexer->handle,UDM_LOG_ERROR,"Error in aliased URL: '%s'",str1);
				freeurl(&curURL);
				FreeDoc(Doc);
				return(IND_OK);
			}else{
				udm_log(Indexer->handle,EXTRA,"Alias: '%s'",str1);
			}
			break;
		}
	}


	/* Compose HTTP header */
	sprintf(str,"%s %s%s HTTP/1.0\r\n",
		(Method==F_HEAD)?"HEAD":"GET",
		CurSrv->proxy?Doc->url:realURL.path,
		CurSrv->proxy?"":realURL.filename);

	if((Doc->status)&&(last_modified[0])&&(!reindex))
		sprintf(STREND(str),"If-Modified-Since: %s\r\n",
			last_modified);

	if((!strcmp(realURL.schema,"http"))||(!strcmp(realURL.schema,"ftp"))){

		/* User agent */
		sprintf(STREND(str),"User-Agent: %s\r\n",UserAgent);

		/* If LocalCharset specified */
		if(LocalCharset)
			sprintf(STREND(str),"Accept-charset: %s\r\n",
				LocalCharset);
		/* Host Name for virtual hosts */
		sprintf(STREND(str),"Host: %s\r\n",realURL.hostname);

		/* Auth if required */
		if(CurSrv->basic_auth)
			sprintf(STREND(str),"Authorization: Basic %s\r\n",
				CurSrv->basic_auth);
	}

	/* Add user defined headers */
	if(*ExtraHeaders)strcat(str,ExtraHeaders);

	/* Empty line is the end of HTTP header */
	strcat(str,"\r\n");

	size=UDM_NET_UNKNOWN;

	THREADINFO(Indexer->handle,"Getting",Doc->url);

	if(Indexer->action==IND_TERMINATED)result=IND_TERMINATED;

	if (use_mirror >= 0) {
		char errstr[STRSIZ]="";
		/* on u_m==0 it returned by mtime from mirrorget */
		/* but we knew that it should be put in mirror  */

		size = MirrorGET(realURL.schema, realURL.hostname,
				realURL.path, realURL.filename,
				Indexer->buf, MaxDocSize,use_mirror,errstr);
		if(size>0){
			udm_log(Indexer->handle,UDM_LOG_DEBUG,"%s has been taken from mirror",Doc->url);
			found_in_mirror=1;
		}else{
			udm_log(Indexer->handle,UDM_LOG_DEBUG,"%s",errstr);
		}
	}


	if(!found_in_mirror){
#ifdef USE_HTDB
		if(!strcmp(realURL.schema,"htdb")){
			size = HTDBGet(Indexer,realURL.path,realURL.filename,
				CurSrv->htdb_list,CurSrv->htdb_doc,
				Indexer->buf,MaxDocSize);
			if(DBErrorCode(Indexer->db))result=IND_ERROR;
		}
#endif
#ifdef USE_FILE
		if(!strcmp(realURL.schema,"file")){
			size = FILEGet(str,Indexer->buf,MaxDocSize);
		}
#endif
#ifdef USE_NEWS
		if((!strcmp(realURL.schema,"news"))){
			size = NEWSGet(realURL.hostname,realURL.port?realURL.port:119,
				str,Indexer->buf,MaxDocSize,CurSrv->read_timeout);
		}else
		if((!strcmp(realURL.schema,"nntp"))){
			size = NNTPGet(realURL.hostname,realURL.port?realURL.port:119,
				str,Indexer->buf,MaxDocSize,CurSrv->read_timeout);
		}
#endif
#ifdef USE_HTTP
		if((!strcmp(realURL.schema,"http"))){
			size = HTTPGet(
				CurSrv->proxy?CurSrv->proxy:realURL.hostname,
				CurSrv->proxy?CurSrv->proxy_port:(realURL.port?realURL.port:80),
				str,Indexer->buf,MaxDocSize,CurSrv->read_timeout);
		}else
		if((!strcmp(realURL.schema,"ftp"))){
			size = HTTPGet(
				CurSrv->proxy?CurSrv->proxy:realURL.hostname,
				CurSrv->proxy?CurSrv->proxy_port:(realURL.port?realURL.port:80),
				str,Indexer->buf,MaxDocSize,CurSrv->read_timeout);
		}
#endif
	}


	if(!result)
	switch(size){
	case UDM_NET_UNKNOWN:
		udm_log(Indexer->handle,WARNING,"Protocol not supported");
		status=HTTP_STATUS_NOT_SUPPORTED;
		result=IND_OK;
		break;
	case UDM_NET_TIMEOUT:
		udm_log(Indexer->handle,WARNING,"Download timeout");
		status=HTTP_STATUS_TIMEOUT;
		result=IND_OK;
		break;
	case UDM_NET_CANT_CONNECT:
		udm_log(Indexer->handle,WARNING,"Can't connect to host %s:%d",
		CurSrv->proxy?CurSrv->proxy:curURL.hostname,
		CurSrv->proxy?CurSrv->proxy_port:(curURL.port?curURL.port:80));
		status=HTTP_STATUS_UNAVAIL;
		result=IND_OK;
		break;
	case UDM_NET_CANT_RESOLVE:
		udm_log(Indexer->handle,WARNING,"Unknown host %s",
			CurSrv->proxy?CurSrv->proxy:curURL.hostname);
		status=HTTP_STATUS_UNAVAIL;
		result=IND_OK;
		break;
	default:
		if(size<0){	/* No connection */
			udm_log(Indexer->handle,WARNING,"Can't connect to host %s:%d",
				CurSrv->proxy?CurSrv->proxy:curURL.hostname,
				CurSrv->proxy?CurSrv->proxy_port:(curURL.port?curURL.port:80));
			status=HTTP_STATUS_UNAVAIL;
			result=IND_OK;
		}else{
			/* Cut response header */
			if((content=strstr(Indexer->buf,"\r\n\r\n"))){
				*content = 0;
				content +=4;
			} else if ((content=strstr(Indexer->buf,"\n\n"))){
				*content = 0;
				content +=2;
			}
			if(!content){
				udm_log(Indexer->handle,UDM_LOG_ERROR,"Illegal HTTP headers in response");
				status=HTTP_STATUS_UNAVAIL;
				result=IND_OK;
			}
		}
	}
	if(Indexer->action==IND_TERMINATED)result=IND_TERMINATED;
	if(result){
		CurSrv->net_errors++;
		freeurl(&curURL);
		freeurl(&realURL);
		if(result!=IND_ERROR)
			result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
		FreeDoc(Doc);
		return(result);
	}

	Indexer->buf[size]=0;
	header=Indexer->buf;
	realsize=size;

	status=changed=1;clone=0;
	content_type=location=href=Indexer->charset=statusline=NULL;
	digest[0]=0;text[0]=0;title[0]=0;keywords[0]=0;
	descript[0]=0;subj[0]=0;last_modified[0]=0;from[0]=0;
	strcpy(last_modified,Doc->last_modified?Doc->last_modified:"");

	Indexer->charset=CurSrv->charset;
	index	=CurSrv->gindex;
	follow	=CurSrv->gfollow;

	size-=(content-Indexer->buf);     /* Could be modified by Content-Length    */
	realsize-=(content-Indexer->buf); /* Will be safe for DIGEST, Parsers etc   */

	THREADINFO(Indexer->handle,"Parsing",Doc->url);
	/* Now lets parse response header lines */
	hcopy = strdup(header);
	s=gettoken(header,"\r\n",&lt);
	while(s){
		if(!STRNCASECMP(s,"HTTP/")){
			status=atoi(s+8);
			statusline=s;
		}else
		if (Force1251 && !STRNCASECMP(s,"Server: ")){
			if (sscanf(s+7,"%s",str1)==1)
				if (!STRNCASECMP(str1,"Microsoft")||
					!STRNCASECMP(str1,"IIS"))
					Indexer->charset="windows-1251";
		}else
		if(!STRNCASECMP(s,"Content-Type: ")){
			content_type=s+14;
			if((Indexer->charset=strstr(content_type,"charset=")))
				Indexer->charset+=8;
		}else
		if(!STRNCASECMP(s,"Location: ")){
			location=s+10;
		}else
		if(!STRNCASECMP(s,"Content-Length: ")){
			size=atoi(s+16);
		}else
		if(!STRNCASECMP(s,"Subject: ")){
			STRNCPY(title,s+9);strcpy(subj,title);
		}else
		if(!STRNCASECMP(s,"From: ")){
			STRNCPY(from,s+6);
		}else
		if(!STRNCASECMP(s,"Newsgroups: ")){
			STRNCPY(keywords,s+12);
		}else
		if(!STRNCASECMP(s,"Last-Modified: ")){
			STRNCPY(last_modified,s+15);
			escstr(last_modified_escaped,last_modified);
		}
		udm_log(Indexer->handle,UDM_LOG_DEBUG,"%s",s);
		s=gettoken(NULL,"\r\n",&lt);
	}
	udm_log(Indexer->handle,EXTRA,"%s %s %d",statusline?statusline:"?",content_type?content_type:"?",size);


	switch(HTTPResponseType(status)){
	case 1: /* No HTTP code */
		CurSrv->net_errors++;
		udm_log(Indexer->handle,UDM_LOG_ERROR,"No HTTP response status");
		result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
		break;

	case HTTP_STATUS_OK:
		if(!content_type){
			udm_log(Indexer->handle,UDM_LOG_ERROR,"No Content-type in '%s'!",Doc->url);
			CurSrv->net_errors++;
			result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
		}
		break;

	case HTTP_STATUS_REDIRECT: /* We'll try to use Location: xxx instead */

		if((Doc->hops<CurSrv->maxhops)&&(location)){
			if((MatchedFilter=WhatToDo(location)))
				Method=MatchedFilter->filter_type;
			else
				Method=F_ALLOW;
			if(Method!=F_DISALLOW){
				if(!parseurl(&newURL,location)){
					if(FindServer(location))
						AddHref(Indexer,location,Doc->url_id,Doc->hops+1,0);
					freeurl(&newURL);
				}else{
					udm_log(Indexer->handle,UDM_LOG_ERROR,"Error in redirect URL: '%s'",location);
				}
			}
		}
		result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
		break;

	case HTTP_STATUS_NOT_MODIFIED:  /* Not Modified, nothing to do */
		result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
		break;

	case HTTP_STATUS_DELETE:
		/* delete it if not robots.txt */
		udm_log(Indexer->handle,EXTRA,"Deleting URL");
		if(!strcmp(curURL.filename,"robots.txt")){
			result=DeleteRobotsFromHost(Indexer,curURL.hostinfo);
			if(result==IND_OK)result=LoadRobots(Indexer);
		}
		if(result!=IND_ERROR){
			if(CurSrv->deletebad){
				result=DeleteUrl(Indexer,Doc->url_id);
			}else{
				result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
			}
		}
		break;

	case HTTP_STATUS_RETRY: /* We'll retry later, maybe host is down */
		CurSrv->net_errors++;
		udm_log(Indexer->handle,EXTRA,"Could not read URL");
		result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
		break;

	default: /* Unknown status, retry later */
		udm_log(Indexer->handle,WARNING,"HTTP %d We don't yet know how to handle it, skipped",status);
		result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
	}
	if(Indexer->action==IND_TERMINATED)result=IND_TERMINATED;
	if(result){
		freeurl(&curURL);
		freeurl(&realURL);
		FreeDoc(Doc);
		return(result);
	}

	/* Now we have HTTP_OK and know Content-Type */

	if ((use_mirror>=0)&&(!found_in_mirror)){
		char errstr[STRSIZ]="";
		/* I tried to get it from mirror, but nothing was there  */
		/* or it was expired. Now we will store it on the mirror */
		if(MirrorPUT(realURL.schema, realURL.hostname,
				realURL.path, realURL.filename,
				hcopy, content, size,errstr)){
			udm_log(Indexer->handle,UDM_LOG_DEBUG,"%s",errstr);
		}
	}
	free_it(hcopy);

#ifdef USE_PARSER
	/* Let's try to start external parser for this Content-Type */
	if((mime=mime_parse(content_type,&mimeno,content,realsize,MaxDocSize-(content-Indexer->buf)))){
		content=mime;content_type=mime_name(mimeno,MIME_TO);
		if((mime=mime_charset(mimeno)))Indexer->charset=mime;
	};
#endif

	/* robots.txt */
	if(!STRNCASECMP(content_type,"text/plain")&&
	(!strcmp(curURL.filename,"robots.txt"))){
		result=ParseRobots(Indexer,content,curURL.hostinfo);
		if(result!=IND_ERROR)result=LoadRobots(Indexer);
		if(result==IND_ERROR){
			freeurl(&curURL);
			freeurl(&realURL);
			FreeDoc(Doc);
			return(IND_ERROR);
		}
	}else

	/* plain TEXT or the same */
	if(!STRNCASECMP(content_type,"text/plain")||
		!STRNCASECMP(content_type,"text/tab-separated-values")||
		!STRNCASECMP(content_type,"text/css"))
	{
		recode_string(content,getcharset(Indexer->charset),LCindex);
		MD5_DIGEST(content,realsize,digest);
		changed=strcmp(digest,Doc->crc);
		origin=FindOrigin(Indexer,digest);
		clone=((origin)&&(origin!=Doc->url_id));

		if(((index)&&(!clone)&&(changed))||reindex){
			strncpy(text,content,MAXTITLESIZE-1);
			text[MAXTITLESIZE-1]=0;
			ParseText(Indexer,content,SEPARATOR,CurSrv->bodyweight,1);
		}
	}else

	/* HTML text */
	if(!STRNCASECMP(content_type,"text/html")){
		recode_string(content,getcharset(Indexer->charset),LCindex);
		MD5_DIGEST(content,realsize,digest);
		changed=strcmp(digest,Doc->crc);
		origin=FindOrigin(Indexer,digest);
		clone=((origin)&&(origin!=Doc->url_id));

		if(((index)&&(!clone)&&(changed))||reindex){
			ParseHtml(Indexer,content,SEPARATOR,&curURL,&realURL,Doc,
				CurSrv,index,follow,text,keywords,
					descript,title);
		}
	}else{
		/* Unknown Content-Type */
		index=0;
		MD5_DIGEST(content,realsize,digest);
		changed=strcmp(digest,Doc->crc);
		origin=FindOrigin(Indexer,digest);
		clone=((origin)&&(origin!=Doc->url_id));
	}
	if((CurSrv->urlweight)&&(strcmp(curURL.filename,"robots.txt"))){
		strcpy(str,Doc->url);
		ParseText(Indexer,str,SEPARATOR,CurSrv->urlweight,1);
	}

	THREADINFO(Indexer->handle,"Updating",Doc->url);
	if(clone){
		udm_log(Indexer->handle,EXTRA,"Duplicate Document with #%d",origin);
		result=DeleteWordFromURL(Indexer,Doc->url_id);
		if(result==IND_OK)
			result=UpdateClone(Indexer,Doc->url_id,status,CurSrv->period,
				content_type,last_modified_escaped,digest);
	}else
	if((!changed)&&(!reindex)){
		result=UpdateUrl(Indexer,Doc->url_id,status,CurSrv->period);
	}else{
		if(subj[0]){
			strcpy(str,subj);
			ParseText(Indexer,str,SEPARATOR,CurSrv->titleweight,1);
		}
		if((result=StoreWords(Indexer,Doc->url_id))==IND_OK)
			result=StoreHrefs(Indexer);

		if(result!=IND_ERROR){
			int total,i;
			char str[STRSIZ]="";
			char lang[3]="";
			
			/* Prepare text,keywords,description */
			if(!strcmp(realURL.schema,"news"))strcpy(keywords,from);

			s=text;while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
			text[MAXTEXTSIZE]=0;escstr(text_escaped,text);

			s=keywords;while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
			keywords[MAXKEYWORDSIZE]=0;escstr(keywords_escaped,keywords);

			s=descript;while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
			descript[MAXDESCSIZE]=0;escstr(descript_escaped,descript);

			s=title;while(*s){if(strchr("\t\n\r",*s))*s=' ';s++;}
			title[MAXTITLESIZE]=0;escstr(title_escaped,title);
			
			qsort((void*)(Indexer->lang+1),MAXLANGPERDOC-1,
				sizeof(UDM_LANG),cmplang);

			total=0;
			for(i=1;i<MAXLANGPERDOC;i++)
				total+=Indexer->lang[i].count;

			if(total){
				sprintf(str,"total: %d",total);
				for(i=1;i<MAXLANGPERDOC;i++){
					if(Indexer->lang[i].count)
						sprintf(STREND(str)," %s:%d (%d%%)",
						Indexer->lang[i].lang,
						Indexer->lang[i].count,
						(int)(Indexer->lang[i].count*100/total));
				}
				udm_log(Indexer->handle,UDM_LOG_DEBUG,"language: %s",str);
				if(Indexer->lang[1].count>10){
					strncpy(lang,Indexer->lang[1].lang,2);
					lang[2]=0;
				}
			}
			

			result=LongUpdateUrl(Indexer,Doc->url_id,status,changed,size,CurSrv->period,
				CurSrv->hint,index,last_modified_escaped,
				text_escaped,title_escaped,
				content_type,keywords_escaped,
				descript_escaped,digest,lang);
		}
	}
	FreeWords(Indexer);
	freeurl(&curURL);
	freeurl(&realURL);
	FreeDoc(Doc);
	return(result);
}

__INDLIB__ INDEXER  * AllocIndexer(int handle){
INDEXER * result;
	result=(INDEXER*)malloc(sizeof(INDEXER));
	result->mwords=0;
	result->nwords=0;
	result->handle=handle;
	result->action=0;
	result->state[0]=0;
	result->Word=NULL;
	result->buf=NULL;
	result->charset=NULL;
	result->db=AllocDB();
	return(result);
}

__INDLIB__ void FreeIndexer(INDEXER *Indexer){

	if(Indexer){
		FreeWords(Indexer);
		free_it(Indexer->Word);
		free_it(Indexer->buf);
		free_it(LocalCharset);
		FreeDB(Indexer->db);
		free(Indexer);
	}
}


__INDLIB__ int ClearDatabase(INDEXER * Indexer){

	return(ClearDB(Indexer));
}

__INDLIB__ char * IndexerErrorMsg(INDEXER *Indexer){
	return(DBErrorMsg(Indexer->db));
}

__INDLIB__ void SetAction(INDEXER *Indexer,int action){
	Indexer->action=action;
}
