#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include "string.h"
#include "common.h"
#include "indexer.h"
#include "dbms.h"
#include "spell.h"
#include "charset.h"
#include "log.h"
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif


#if (WIN32|WINNT)
#include <windows.h>
#ifdef HAVE_PTHREAD
static CRITICAL_SECTION main_mutex;
#endif
#else
#include <unistd.h>
#ifdef HAVE_PTHREAD
#include <pthread.h>
static pthread_mutex_t main_mutex;
#endif
#endif


static int seconds=0;
static int nthreads=0;
static index_flags=0;
#ifdef HAVE_PTHREAD
static int thread_number=1;
#else
static int thread_number=0;
#endif

static char *config_name= CONFIGDIR "/indexer.conf";


int usage(){

	fprintf(stderr,
"\n\
indexer from UdmSearch v." VERSION "/%s\n\
http://mysearch.udm.net (C) 1998-1999, UdmSearch Developers Team.\n\
\n\
Usage: indexer [OPTIONS]  [configfile]\n\
\n\
Indexing options:\n\
  -a            reindex All documents (even if not expired)\n\
  -m            reindex if not Modified\n\
  -e            more Expired first\n\
  -n number     index only Number documents and exit\n\
  -p number     sleeP number seconds after each URL\n\
  -q            quick startup (do not add Server URLs)\n\
"
#ifdef HAVE_PTHREAD
"  -N number     run number threads\n\
"
#endif
"\n\
Subsection control options:\n\
  -t tag        set filter on Tag\n\
  -u pattern    set filter on URL (SQL LIKE wildcards)\n\
  -s status     set filter on HTTP Status\n\
\n\
Logging options:\n\
"
#ifdef LOG_PERROR
"  -l            do not Log to stdout/stderr\n\
"
#endif
"  -v number     Verbose level, 0-5\n\
\n\
Mics. options:\n\
  -C            Clear database and exit\n\
  -S            print Statistics and exit\n\
  -I            print referers and exit\n\
  -h,-?         print this Help page and exit\n\
\n\
\n\
Mail bug reports and suggestions to <devel@search.udm.net>.\n",DBVersion);

	return(0);
}



static int IndexerMain(INDEXER * Indexer){
int res=0;

	while(1){
		res=IndexNextURL(Indexer,index_flags);
		if(seconds){
			udm_log(Indexer->handle,UDM_LOG_DEBUG,"Sleeping %d second(s)",seconds);
			UDMSLEEP(seconds);
		}
		switch(res){
			case IND_OK:
				continue;				
			case IND_ERROR:
				THREADINFO(Indexer->handle,"Error",IndexerErrorMsg(Indexer));
				udm_log(Indexer->handle,UDM_LOG_ERROR,"Error: '%s'",IndexerErrorMsg(Indexer));
				return(1);
			case IND_NO_TARGET:
#ifdef HAVE_PTHREAD

				INC_ACC();

				LOCK_MAIN();
				UDMSLEEP(1);
				res=(GetActive()>=nthreads);
				UNLOCK_MAIN();

				DEC_ACC();
				if(res)return(0);
				continue;
#else
				return(0);
#endif
			default:
				return(0);
		}
	}
	return(0);
}


INDEXER * I[100];
#if  (WIN32|WINNT)
DWORD WINAPI thread_main(void *arg){
#else
void * thread_main(void *arg){
#endif
int mynum;

	LOCK_MAIN();
	mynum=thread_number++;
	I[mynum]=AllocIndexer(mynum);
	UNLOCK_MAIN();

	IndexerMain(I[mynum]);

	LOCK_MAIN();
	nthreads--;
	udm_log(I[mynum]->handle,INFO,"Done");
	FreeIndexer(I[mynum]);
	I[mynum]=0;
	UNLOCK_MAIN();

	return(NULL);
}


int main(int argc, char **argv) {

char *language=NULL;
int npages=-1;
int reindexall=0,clear=0,stat=0,integrity=0;
int log2stderr=1,dump=0;
int add_servers=1;
extern char *optarg;
extern int optind;
int ch;
int maxthreads=1;

	while ((ch = getopt(argc, argv, "CSIahelmdq?t:u:s:n:v:L:A:D:p:N:")) != -1){
		switch (ch) {
		case 'C': clear++;add_servers=0;break;	
		case 'S': stat++;add_servers=0;break;
		case 'I': integrity++;add_servers=0;break;
		case 'L': language=optarg;break;
		case 'q': add_servers=0;break;
		case 'l': log2stderr=0;break;
		case 'a': reindexall=1;break;
		case 'e': index_flags|=FLAG_EXP_FIRST;break;
		case 'm': index_flags|=FLAG_REINDEX;break;
		case 'n': npages=atoi(optarg);break;
		case 'v': SetLogLevel(atoi(optarg));break;
		case 'p': seconds=atoi(optarg);break;
		case 'd': dump=1;break;
		case 't': AddTagLimit(atoi(optarg));break;
		case 's': AddStatusLimit(atoi(optarg));break;
		case 'u': AddURLLimit(optarg);break;
		case 'N': maxthreads=atoi(optarg);break;
		case '?':
		case 'h':
		default:  
			usage();
			return(1);
		}
	}

	argc -= optind;argv += optind;

	if(argc>1){
		usage();
		return(1);
	}
	if(argc==1)config_name=argv[0];

#if (WIN32|WINNT)
#else
	udm_openlog(log2stderr);
#endif



/*START:*/


	InitCharset();

#ifdef HAVE_PTHREAD
	InitMutexes();
#if (WIN32|WINNT)
	InitializeCriticalSection(&main_mutex);
#else
	
	pthread_mutex_init(&main_mutex,NULL);
#endif
#endif

	

	if(LoadConfig(config_name,0,
		(add_servers?FLAG_ADD_SERV:0)+
		(reindexall?FLAG_MARK:0)+
		((!integrity&&!stat&&!clear)*FLAG_SPELL)
		))
			exit(1);

	if(clear){
		INDEXER * Indexer;
#ifndef DEBUG
		char str[5]="";
		printf("You are going to delete database '%s' content\n",DBName?DBName:"");
		printf("Are you sure?(YES/no)");
		if(fgets(str,sizeof(str),stdin)){
			if(strncmp(str,"YES",3))
				printf("Canceled\n");
			else{
#endif
				Indexer=AllocIndexer(0);
				if(IND_OK!=ClearDatabase(Indexer)){
					udm_log(Indexer->handle,UDM_LOG_ERROR,"Error: '%s'",IndexerErrorMsg(Indexer));
				}
				FreeIndexer(Indexer);
#ifndef DEBUG
			}
		}else
			printf("Canceled\n");
#endif

	}else
	if(stat){
		INDEXER * Indexer;
		Indexer=AllocIndexer(0);
		if(IND_OK!=ShowStatistics(Indexer)){
			udm_log(Indexer->handle,UDM_LOG_ERROR,"Error: '%s'",IndexerErrorMsg(Indexer));
		}
		FreeIndexer(Indexer);
	}else
	if(integrity){
		INDEXER * Indexer;
		Indexer=AllocIndexer(0);
		if(IND_OK!=ShowReferers(Indexer)){
			udm_log(Indexer->handle,UDM_LOG_ERROR,"Error: '%s'",IndexerErrorMsg(Indexer));
		}
		FreeIndexer(Indexer);
	}else{
#if (WIN32|WINNT)
#else
		udm_log(0,INFO, "indexer from UdmSearch v." VERSION"/%s started with '%s'", DBVersion, config_name);
#endif

		SetMaxURLNumber(npages);

#ifdef HAVE_PTHREAD
		{
#if (WIN32|WINNT)
#else
			pthread_t *thread;
#endif
			int i;
			thread_number=1;
			for(i=0;i<maxthreads;i++){
#if (WIN32|WINNT)
#else
				thread=(pthread_t*)malloc(sizeof(pthread_t));
#endif
				LOCK_MAIN();
				nthreads++;
				UNLOCK_MAIN();
#if (WIN32|WINNT)
				CreateThread(NULL, 0, &thread_main, NULL, 0, NULL);
#else
				pthread_create(thread,NULL,&thread_main,NULL);
#endif
			}
		}
		
		while(1){
			LOCK_MAIN();
			if(!nthreads){
				UNLOCK_MAIN();
				break;
			}
			UNLOCK_MAIN();
			UDMSLEEP(1);
		}
		DestroyMutexes();

#if (WIN32|WINNT)
		DeleteCriticalSection(&main_mutex);
#else
		pthread_mutex_destroy(&main_mutex);
#endif
#else
		thread_main(NULL);
#endif
	}
	thread_number=0;

//goto START;

	return(0);
}
